@@ -8,18 +8,18 @@
  * Copyright (C) 1996 Paul Mackerras.
  */
 #include <linux/config.h>
-#include <linux/module.h>
 #include <linux/kernel.h>
 #include <linux/delay.h>
 #include <linux/types.h>
 #include <linux/string.h>
-#include <linux/malloc.h>
+#include <linux/slab.h>
 #include <linux/blk.h>
 #include <linux/proc_fs.h>
 #include <linux/stat.h>
 #include <linux/tqueue.h>
 #include <linux/interrupt.h>
 #include <linux/reboot.h>
+#include <linux/spinlock.h>
 #include <asm/dbdma.h>
 #include <asm/io.h>
 #include <asm/pgtable.h>
@@ -28,11 +28,10 @@
 #include <asm/irq.h>
 #include <asm/hydra.h>
 #include <asm/processor.h>
-#include <asm/spinlock.h>
 #include <asm/feature.h>
 #ifdef CONFIG_PMAC_PBOOK
-#include <asm/adb.h>
-#include <asm/pmu.h>
+#include <linux/adb.h>
+#include <linux/pmu.h>
 #endif
 
 #include "scsi.h"
@@ -70,11 +69,6 @@ unsigned char use_active_neg = 0;  /* bit mask for SEQ_ACTIVE_NEG if used */
 #define ALLOW_DEBUG(tgt)	((mesh_debug_targets >> (tgt)) & 1)
 #define DEBUG_TARGET(cmd)	((cmd) && ALLOW_DEBUG((cmd)->target))
 
-struct proc_dir_entry proc_scsi_mesh = {
-	PROC_SCSI_MESH, 4, "mesh",
-	S_IFDIR | S_IRUGO | S_IXUGO, 2
-};
-
 #undef MESH_DBG
 #define N_DBG_LOG	50
 #define N_DBG_SLOG	20
@@ -159,8 +153,9 @@ struct mesh_state {
 	struct dbdma_cmd *dma_cmds;	/* space for dbdma commands, aligned */
 	int	clk_freq;
 	struct mesh_target tgts[8];
-	struct device_node *dnode;
-	unsigned char *mio_base;
+	void	*dma_cmd_space;
+	struct device_node *ofnode;
+	u8*	mio_base;
 #ifndef MESH_NEW_STYLE_EH
 	Scsi_Cmnd *completed_q;
 	Scsi_Cmnd *completed_qtail;
@@ -173,14 +168,6 @@ struct mesh_state {
 #endif
 };
 
-#ifdef CONFIG_PMAC_PBOOK
-static int mesh_notify_sleep(struct pmu_sleep_notifier *self, int when);
-static struct pmu_sleep_notifier mesh_sleep_notifier = {
-	mesh_notify_sleep,
-	SLEEP_LEVEL_BLOCK,
-};
-#endif
-
 #ifdef MESH_DBG
 
 static void dlog(struct mesh_state *ms, char *fmt, int a);
@@ -230,6 +217,14 @@ static int data_goes_out(Scsi_Cmnd *);
 static void do_abort(struct mesh_state *ms);
 static void set_mesh_power(struct mesh_state *ms, int state);
 
+#ifdef CONFIG_PMAC_PBOOK
+static int mesh_notify_sleep(struct pmu_sleep_notifier *self, int when);
+static struct pmu_sleep_notifier mesh_sleep_notifier = {
+	mesh_notify_sleep,
+	SLEEP_LEVEL_BLOCK,
+};
+#endif
+
 static struct notifier_block mesh_notifier = {
 	mesh_notify_reboot,
 	NULL,
@@ -264,6 +259,7 @@ mesh_detect(Scsi_Host_Template *tp)
 		mesh = find_compatible_devices("scsi", "chrp,mesh0");
 	for (; mesh != 0; mesh = mesh->next) {
 		struct device_node *mio;
+		
 		if (mesh->n_addrs != 2 || mesh->n_intrs != 2) {
 			printk(KERN_ERR "mesh: expected 2 addrs and 2 intrs"
 			       " (got %d,%d)", mesh->n_addrs, mesh->n_intrs);
@@ -275,7 +271,7 @@ mesh_detect(Scsi_Host_Template *tp)
 			continue;
 		}
 		mesh_host->unique_id = nmeshes;
-#ifndef MODULE
+#if !defined(MODULE)
 		note_scsi_host(mesh, mesh_host);
 #endif
 
@@ -284,7 +280,7 @@ mesh_detect(Scsi_Host_Template *tp)
 			panic("no mesh state");
 		memset(ms, 0, sizeof(*ms));
 		ms->host = mesh_host;
-		ms->dnode = mesh;
+		ms->ofnode = mesh;
 		ms->mesh = (volatile struct mesh_regs *)
 			ioremap(mesh->addrs[0].address, 0x1000);
 		ms->dma = (volatile struct dbdma_regs *)
@@ -301,6 +297,7 @@ mesh_detect(Scsi_Host_Template *tp)
 		ms->dma_cmds = (struct dbdma_cmd *) DBDMA_ALIGN(dma_cmd_space);
 		memset(ms->dma_cmds, 0, (mesh_host->sg_tablesize + 1)
 		       * sizeof(struct dbdma_cmd));
+		ms->dma_cmd_space = dma_cmd_space;
 
 		ms->current_req = 0;
 		for (tgt = 0; tgt < 8; ++tgt) {
@@ -329,15 +326,13 @@ mesh_detect(Scsi_Host_Template *tp)
 			mesh_sync_period = minper;
 
 		ms->mio_base = 0;
-		for (mio = ms->dnode->parent; mio; mio = mio->parent)
-			if (strcmp(mio->name, "mac-io") == 0
-			    && mio->n_addrs > 0)
+		for (mio = ms->ofnode->parent; mio; mio = mio->parent)
+			if (strcmp(mio->name, "mac-io") == 0 && mio->n_addrs > 0)
 				break;
 		if (mio)
-			ms->mio_base = (unsigned char *)
-				ioremap(mio->addrs[0].address, 0x40);
-
+			ms->mio_base = (u8 *) ioremap(mio->addrs[0].address, 0x40);
 		set_mesh_power(ms, 1);
+
 		mesh_init(ms);
 
 		if (request_irq(ms->meshintr, do_mesh_interrupt, 0, "MESH", ms)) {
@@ -351,27 +346,45 @@ mesh_detect(Scsi_Host_Template *tp)
 #ifdef CONFIG_PMAC_PBOOK
 		pmu_register_sleep_notifier(&mesh_sleep_notifier);
 #endif /* CONFIG_PMAC_PBOOK */
- 		register_reboot_notifier(&mesh_notifier);
-	}		
+		register_reboot_notifier(&mesh_notifier);
+	}
 
 	return nmeshes;
 }
 
+int
+mesh_release(struct Scsi_Host *host)
+{
+	struct mesh_state *ms = (struct mesh_state *) host->hostdata;
+
+	if (ms == 0)
+		return 0;
+	if (ms->mesh)
+		iounmap((void *) ms->mesh);
+	if (ms->dma)
+		iounmap((void *) ms->dma);
+	if (ms->mio_base)
+		iounmap((void *) ms->mio_base);
+	kfree(ms->dma_cmd_space);
+	free_irq(ms->meshintr, ms);
+	feature_clear(ms->ofnode, FEATURE_MESH_enable);
+	return 0;
+}
+
 static void
 set_mesh_power(struct mesh_state *ms, int state)
 {
-	if (_machine != _MACH_Pmac || ms->mio_base == 0)
+	if (_machine != _MACH_Pmac)
 		return;
-		
 	if (state) {
-		feature_set(ms->dnode, FEATURE_MESH_enable);
+		feature_set(ms->ofnode, FEATURE_MESH_enable);
 		/* This seems to enable the termination power. strangely
 		   this doesn't fully agree with OF, but with MacOS */
 		if (ms->mio_base)
 			out_8(ms->mio_base + 0x36, 0x70);
 		mdelay(200);
 	} else {
-		feature_clear(ms->dnode, FEATURE_MESH_enable);
+		feature_clear(ms->ofnode, FEATURE_MESH_enable);
 		if (ms->mio_base)
 			out_8(ms->mio_base + 0x36, 0x34);
 		mdelay(10);
@@ -413,7 +426,6 @@ mesh_notify_sleep(struct pmu_sleep_notifier *self, int when)
 	return PBOOK_SLEEP_OK;
 }
 #endif /* CONFIG_PMAC_PBOOK */
- 
 
 int
 mesh_queue(Scsi_Cmnd *cmd, void (*done)(Scsi_Cmnd *))
@@ -1912,12 +1924,6 @@ data_goes_out(Scsi_Cmnd *cmd)
 	}
 }
 
-#ifdef MODULE
-Scsi_Host_Template driver_template = SCSI_MESH;
-
-#include "scsi_module.c"
-#endif /* MODULE */
-
 #ifdef MESH_DBG
 static inline u32 readtb(void)
 {
@@ -2008,3 +2014,7 @@ static void dumpslog(struct mesh_state *ms)
 	} while (i != ms->log_ix);
 }
 #endif /* MESH_DBG */
+
+static Scsi_Host_Template driver_template = SCSI_MESH;
+
+#include "scsi_module.c"
